package org.integrade.asct.ws;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import javax.security.auth.callback.CallbackHandler;

import org.integrade.asct.ws.datatypes.Application;
import org.integrade.asct.ws.datatypes.Execution;

import asct.core.corba.AsctImpl;
import asct.core.corba.OrbHolder;
import asct.core.repository.ApplicationRepositoryManager;
import asct.core.submission.ExecutionRequestManager;
import asct.shared.AbstractGridApplication;
import asct.shared.ExecutionRequestData;
import asct.shared.ExecutionRequestStatus;
import asct.shared.IExecutionListener;
import br.usp.ime.oppstore.broker.OppStoreBroker;
import clusterManagement.ApplicationNotFoundException;
import clusterManagement.ApplicationRegistrationException;
import clusterManagement.BinaryCreationException;
import clusterManagement.BinaryNotFoundException;
import clusterManagement.ContextInitiationException;
import clusterManagement.DirectoryCreationException;
import clusterManagement.DirectoryNotEmptyException;
import clusterManagement.DirectoryNotFoundException;
import clusterManagement.InvalidPathNameException;
import clusterManagement.SecurityException;
import dataTypes.ApplicationDescription;
import dataTypes.ContentDescription;

public class WSApplicationFacade {
	
	
	/** Manages the execution requests. */
	private ExecutionRequestManager execManager;

	/** Manages the interaction with the application repository. */
	private ApplicationRepositoryManager appRepManager;

	/** Allows accessing the OppStore broker through a JNI interface. */
	private OppStoreBroker broker;
	
	private static WSApplicationFacade instance;
	
	private static void init() {
		Properties asctProperties = new Properties();
		boolean isSecure = false, useOppStore = false;
		try {
			asctProperties.load(new FileInputStream("asct.conf"));
			String isSecurityEnabled = asctProperties
					.getProperty("isSecurityEnabled");
			String isOppStoreEnabled = asctProperties
					.getProperty("isOppStoreEnabled");

			if (isSecurityEnabled != null
					&& isSecurityEnabled.equalsIgnoreCase("true")) {
				System.out.println("[ASCTController] Security is enabled.");
				isSecure = true;
			}

			if (isOppStoreEnabled != null
					&& isOppStoreEnabled.equalsIgnoreCase("true")) {
				System.out.println("[ASCTController] OppStore is enabled.");
				useOppStore = true;
			}
			
			cleanOutputDirectory();
			
		} catch (IOException e) {
			System.err.println("[ASCTController] WARNING: could not find 'asct.conf'. Using default values.");
		}
		instance = new WSApplicationFacade(getOutputDirectory(), isSecure, useOppStore, null);
	}
	
	private static void cleanOutputDirectory() {
		File outputDirectory = new File(getOutputDirectory());
		if (outputDirectory.isDirectory()) {
			String[] children = outputDirectory.list();
			for (int i = 0; i < children.length; i++) {
				boolean success = deleteDir(new File(outputDirectory,
						children[i]));
				if (!success) {
					System.out.println("Error cleaning the output directory");
				}
			}
		}
	}
	
	private static boolean deleteDir(File dir) {
		if (dir.isDirectory()) {
			String[] children = dir.list();
			for (int i = 0; i < children.length; i++) {
				boolean success = deleteDir(new File(dir, children[i]));
				if (!success) {
					return false;
				}
			}
		}
		// The directory is now empty so delete it
		return dir.delete();
	}

	public static String getOutputDirectory() {
		return new String(System.getProperty("user.home") + "/output/");
	}
	
	public static WSApplicationFacade getInstance() {		
		
		if (instance == null) {
			init();
		}
		return instance;
	}
	
	/**
	 * Constructor.
	 * */
	private WSApplicationFacade(String localDirectory,boolean isSecure, boolean useOppStore, CallbackHandler securityCallbackHandler) {
		OrbHolder orbHolder = new OrbHolder();
		orbHolder.initStubs();
		
		if (useOppStore == true){
			broker = new OppStoreBroker();
		}
		
		execManager = new ExecutionRequestManager(orbHolder, localDirectory, broker);		
		
		try {

			appRepManager = new ApplicationRepositoryManager(
					orbHolder.getApplicationRepositoryStubWrapper(isSecure,securityCallbackHandler));
			
		} catch (ContextInitiationException e1) {
			// TODO Ricardo 
			e1.printStackTrace();
		}
		
		orbHolder.setAsctImpl(new AsctImpl(execManager));
		
		new Thread(orbHolder).start();
		
		while (orbHolder.getIsRunning() == false) {
			try { Thread.sleep(100);} 
			catch (InterruptedException e) {}
		}
	}


	/** ************************** Submission **************************** */

	/**
	 * @param data all data needed for submission
	 * @return new request status or null if unknown status
	 * @throws SecurityException 
	 */
	public ExecutionRequestStatus executeApplication(
				final ExecutionRequestData data) 
			throws ApplicationNotFoundException, DirectoryNotFoundException, 
				InvalidPathNameException, SecurityException {
		data.getApplication().setDescription(appRepManager.getAppDescription(data.getApplication()));
		
		return execManager.executeApplication(data);
	}

	/**
	 * @param data all data needed for submission
	 * @param binaryFileName name of the binary to be submitted
	 * @return new request status
	 */
	public ExecutionRequestStatus executeBinary(
				final ExecutionRequestData data, final String binaryFileName) {
		return execManager.executeBinary(data, binaryFileName);
	}

	/**
	 * @param requestId id of the request to be killed
	 */
	public void killApplication(final String requestId){
		execManager.killApplication(requestId);
	}

	/**
	 * @param requestId id of the request to be killed
	 */
	public boolean killApplicationExecution(final String requestId){
		if (execManager.executionIdIsValid(requestId)){
			execManager.killApplication(requestId);
			return true;
		}
		else
			return false;
	}
	
	/**
	 * 
	 * @param execReqId id of the request to get results
	 */
	public void getApplicationResults(final String execReqId, String[] outputFiles) {
		execManager.getApplicationResults(execReqId, outputFiles);
	}
	
	/** ******************** Application Repository ************************ */

	/**
	 * @return contents of the application repository root directory
	 * @throws SecurityException 
	 * */
	public ContentDescription[] listRootDirectoryContents() 
			throws DirectoryNotFoundException, InvalidPathNameException, SecurityException{
		return appRepManager.listRootDirectoryContents();
	}

	
	/**
	 * @param directoryPath the relative path from the repository root 
	 * @return contents of the directory in the application repository
	 * @throws SecurityException 
	 */
	 public ContentDescription[] listDirectoryContents(
			 final String directoryPath) 
	 		throws DirectoryNotFoundException, InvalidPathNameException, SecurityException { 
		 return appRepManager.listDirectoryContents(directoryPath);
	 }
	 
	 /**
	  * @param directoryPath the relative path from the repository root 
	 * @throws SecurityException 
	  */
	 public void createDirectory(final String directoryPath) 
	 		throws DirectoryCreationException, InvalidPathNameException, SecurityException {
		 appRepManager.createDirectory(directoryPath);
	 }
	 
	 /**
	  * @param directoryPath the relative path from the repository root 
	 * @throws SecurityException 
	  */
	 public void removeDirectory(final String directoryPath)
	 			throws DirectoryNotFoundException, DirectoryNotEmptyException,
	 				InvalidPathNameException, SecurityException {
		 appRepManager.removeDirectory(directoryPath);
	 }

	 /**
	  * @param basePath the relative path from the repository root 
	  * @param applicationName the name of the application
	 * @throws SecurityException 
	  */
	 public void registerApplication(final String basePath, 
			 		final String applicationName) 
	 			throws ApplicationRegistrationException, 
					DirectoryCreationException, InvalidPathNameException, SecurityException {
		 appRepManager.registerApplication(basePath, applicationName);
	 }
	 
	 /**
	  * @param basePath the relative path from the repository root 
	  * @param applicationName the name of the application
	 * @throws SecurityException 
	  */
	 public void unregisterApplication(final String basePath, 
		 			final String applicationName) 
	 			throws ApplicationNotFoundException, DirectoryNotFoundException,
	 				DirectoryNotEmptyException, InvalidPathNameException, SecurityException {
		 appRepManager.unregisterApplication(basePath, applicationName);
	 }
	 
	/**
	  * @param localFilePath the local path of the file to be uploaded 
	  * @param remoteBasePath the relative path from the repository root 
	  * @param applicationName the name of the application
	  * @param platform os name plus hardware (eg. Linux_i686)
	 * @throws SecurityException 
	 * @throws FileNotFoundException 
	 */
	public void uploadBinary(final String localFilePath, 
				final String remoteBasePath, final String applicationName, 
				final String platform) 
			throws BinaryCreationException, ApplicationNotFoundException,
					DirectoryNotFoundException, InvalidPathNameException, SecurityException, FileNotFoundException {
		System.out.println("Upload "+localFilePath+" rbp "+remoteBasePath+" platform "+platform+" apn "+applicationName);
		appRepManager.uploadBinary(localFilePath, remoteBasePath, 
				applicationName, platform);
	}

	/**
	  * @param basePath the relative path from the repository root 
	  * @param applicationName the name of the application
	  * @param fileName the name of the file to be deleted
	 * @throws SecurityException 
	 */
	public void deleteBinary(final String basePath, 
				final String applicationName, final String fileName) 
			throws ApplicationNotFoundException, DirectoryNotFoundException, 
					BinaryNotFoundException, InvalidPathNameException, SecurityException {
		appRepManager.deleteBinary(basePath, applicationName, fileName);
	}

	/** ******************** OppStore ************************ */

	/**
	 * Upload InputFiles to OppStore
	 * TODO: Must also work with parametric applications
	 */		 
	public String[] uploadFilesToOppStore(String[] inputFiles) {

		if (broker != null) {
			String[] inputKeys = new String[ inputFiles.length ];
			boolean successfulUpload = true;
			for ( int i=0; i < inputFiles.length && successfulUpload ; i++ ) {
				String inputKey = uploadSingleFileToOppStore( inputFiles[i] );			
				if ( inputKey != null ) {				
					String[] splitFileName = inputFiles[i].split("/");
					String fileName = splitFileName[ splitFileName.length - 1]; 
					String integradeKey = fileName + "[key]" + inputKey;
					inputKeys[i] = integradeKey;
					System.out.println("AsctController -> fileName:" + integradeKey + ".");
				}
				else { 
					successfulUpload = false;
					System.out.println("Input file upload failed.");
				}

			}		
			if (successfulUpload) 
				return inputKeys;			
		}
		
		// Files were not successfully uploaded
		return null;
	}

	private String uploadSingleFileToOppStore(String fileName) {

		File file = new File(fileName);
		String key = null;
		if (file.isFile())
			key = broker.storeFileW(fileName, null);

		return key;
	}

	/** ************************** Outros ********************************** */

	/**
	 * 
	 * @param listener a listener to execution requests
	 */
	public void registerExecutionStateListener(
				final IExecutionListener listener) {
		execManager.registerExecutionStateListener(listener);
	}
	
	public ApplicationDescription getApplicationDescription(String name, String basePath) 
		throws ApplicationNotFoundException, DirectoryNotFoundException, 
			   InvalidPathNameException, SecurityException {
		AbstractGridApplication application = new DummyApplication(name, basePath);
		return appRepManager.getAppDescription(application);
	}


	class DummyApplication extends AbstractGridApplication {

		public DummyApplication(String name, String remoteApplicationPath) {
			super(name, remoteApplicationPath);
		}
		
	}


	public Application getApplicationById(String appId)
			throws ApplicationNotFoundException, DirectoryNotFoundException, 
			   InvalidPathNameException, SecurityException {
		ApplicationDescription appDesc = appRepManager.getAppDescriptionById(appId);
		Application app = new Application (appDesc);
		return app;
	}
	
	public ApplicationDescription getApplicationDescriptionById(String appId)
			throws ApplicationNotFoundException, DirectoryNotFoundException, 
			   InvalidPathNameException, SecurityException {
		return appRepManager.getAppDescriptionById(appId);
	}
	
	public List<Application> getApplicationList(String query) throws SecurityException {
		ApplicationDescription[] appDescs = appRepManager.listApplicationDescriptions(query);
		List<Application> applications = new ArrayList<Application>();
		for (ApplicationDescription appDesc : appDescs) {
			Application app = new Application(appDesc);
			applications.add(app);
		}
		return applications;
	}
	

	public List<Execution> getExecutionList(String query) 
			throws SecurityException {
		List<Execution> executions = new ArrayList<Execution>();
		return executions;
	}

	public Execution getExecutionDetails(String id) 
			throws SecurityException {
		Execution execution = new Execution ();

		AbstractGridApplication AGApp;
		AGApp = execManager.getGridApplication(id);
		
		execution.setApplicationId(AGApp.getApplicationDescription().appId);

		execution.setArguments(AGApp.getArguments());
		execution.setApplicationType(AGApp.getApplicationType());
		execution.setConstraints(AGApp.getConstraints());
		execution.setPreferences(AGApp.getPreferences());
		
		// TODO: store and retrieve this data
		execution.setForceDifferentNodes("False");
		execution.setStoreStderr("False");
		execution.setStoreStdout("False");
		execution.setNumberOfCopies(0);
		execution.setNumberOfTasks(0);
		
		return execution;
	}
	
	public ExecutionRequestStatus getExecutionStatus(String id) 
			throws SecurityException {
		return execManager.getExecutionStatus(id);
	}
}
