package moca.core.proxy.receivers;

import moca.core.proxy.ClientInfo;


/**
 * Package  uma classe de controle interno do proxy, que encapsula mensagens
 * e mantem informaes necessrias para que a sequencia das mensagens servidor-proxy
 * seja a mesma que a proxy-cliente.
 * Cada pacote possui dois identificadores, um para o proximo pacote (caso ele exista) 
 * e o outro para o pacote anterior (caso ele exista). Os pacotes tambm possuem uma
 * marcao que permite informar qual o estagio do pacote dentro do proxy. 
 * Primeiramente o pacote recebe a marcao PROCESSING, que indica que ele ainda
 * esta sendo processado em algum dois receivers intermedirios. Uma vez que ele chegue
 * ao Sender (Ultimo estagio do proxy), duas situaes podem ocorrer:
 * 
 * A primeira delas acontece quando o pacote anterior no existe ou j est adicionado 
 * na fila do SENDER. Neste caso, o pacote corrente no s poder tambm ser inserido nesta 
 * fila e receber o marcador READY, como tambm iterar nos pacotes posteriores em busca de 
 * pacotes no estado PROCESSED (estgio de pendencia).
 * 
 * A segunda situao acontece quando o pacote anterior est em um estagio diferente de READY.
 * Neste caso, o pacote no poder ser adicionado a fila do Sender, permanecendo pendente e 
 * com marcao PROCESSED.   
 * 
 *  
 * @author hana
 *
 */
public class Package {
	/**
	* Indica que o pacote est sendo processado em algum receiver
	*/
	private static final int PROCESSING = 1;
	/**
	 * Indica que o pacote foi processado mas ainda no foi adicionado a fila do Sender
	 */
	private static final int PROCESSED = 2;
	/**
	 * Indica que o pacote foi processado e j foi adicionado a fila do Sender
	 */
	private static final int READY = 3;
	/**
	 * ponteiro para o pacote posterior (proximo pacote na ordem de envio)
	 */
	private Package next;
	/**
	 * ponteiro para o pacote anterior 
	 */
	private Package previous;
	/**
	 * O status inicial de um pacote  processing
	 */
	private int status = PROCESSING;
	/**
	 * Mensagem encapsulada pelo package
	 */
	private ReceiverMessage msg;
	/**
	 * Destinatrio da mensagem
	 */
	private ClientInfo info;
	/**
	 * Destino do pacote
	 */
	private Object destination;
	/**
	 * Construtor que recebe um destinatario e uma mensagem
	 * @param info informaes do destinatario
	 * @param msg mesnagem
	 */
	public Package(ClientInfo info, ReceiverMessage msg) {
		this.msg = msg;
		this.info = info;
	}
	/**
	 * Recupera a mensagem encapsulada pelo package
	 * @return mensagem
	 */
	public synchronized ReceiverMessage getMessage() {
		return this.msg;
	}
	/**
	 * Muda a mensagem do package
	 * @param msg mensagem
	 */
	public synchronized void setMessage(ReceiverMessage msg) {
		this.msg = msg;
	}
	/**
	 * Recupera as informacoes do destinatario da mensagem encapsulada 
	 * pelo package
	 * @return destinatario  
	 */
	
	public synchronized ClientInfo getClientInfo() {
		return this.info;
	}
	/**
	 * Seta o pacote anterior
	 * @param previous pacote anterior
	 */
	private synchronized void setPrevious(Package previous) {
		this.previous = previous;
	}
	/**
	 * seta o proximo pacote
	 * @param next
	 */
	private synchronized void setNext(Package next) {
		this.next = next;
	}
	/**
	 * Recupera o status do pacote (PROCESSING, PROCESSED ou READY)
	 * @return
	 */
	private synchronized int getStatus() {
		return this.status;
	}
	/**
	 * Seta o pacote anterior
	 * @param pack
	 */
	public synchronized void setPreviousPackage(Package pack) {
		this.previous = pack;
		if (pack != null) {
			pack.setNext(this);
		}
	}
	/**
	 * Recupera o proximo destino do pacote (proximo receiver)
	 * @return proximo receiver
	 */
	public synchronized Object getDestination() {
		return this.destination;
	}
	/**
	 * define proximo receiver que recebera o pacote
	 * @param obj identificador do receiver
	 */
	public synchronized void setDestination(Object obj) {
		this.destination = obj;
	}
	/**
	 * Seta o pacote como processado
	 */
	public synchronized void setAsProcessed() {
		this.status = PROCESSED;
	}
	/**
	 * Uma mensagem s  enviada se a sua mensagem anterior (da qual ela depende) est pronta
	 * para envio. Isso porque, uma vez na fila de entrega, as mensagens no podem mais mudar 
	 * de ordem
	 * @param disp
	 */
	public void onSender(Sender sender) {
		Package next = null;
		synchronized (this) {
			if (this.status != PROCESSED) {
				//nao h o que fazer
				return;
			}
			next = this.next;
			//this.status == PROCESSED
			if (this.previous == null || this.previous.getStatus() == READY) {
				//pronta para ser entregue
				this.status = READY;
				sender.add(this);
				this.previous = null; //evitar vazamento de memoria
				this.next = null; //evitar vazamento de memoria
			}						
		}
		if (next != null) {
			next.onSender(sender); //verifique se existem dependentes prontas
		}
	}
}
