package moca.core.proxy.state.filter.xml;

import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.List;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import moca.core.proxy.state.filter.BooleanValue;
import moca.core.proxy.state.filter.Filter;
import moca.core.proxy.state.filter.FilterBuilder;
import moca.core.proxy.state.filter.LiteralOperator;
import moca.core.proxy.state.filter.Operator;

import org.apache.log4j.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * Carrega um filtro a partir de um n DOM
 * @author hana
 */
public class DOMFilterBuilder implements FilterBuilder {
	private static final Logger LOGGER =
		Logger.getLogger(DOMFilterBuilder.class);
	/**
	* Constante que armazena o nome do arquivo de propriedades onde esto mapeados
	* os nomes de tags aos nomes dos operadores correspondentes
	*/
	private static final String ELEMENT_OPERATOR_MAP = "ElementOperatorMapping";
	/**
	* N raiz
	*/
	private Element root;
	/**
	 * Arquivo de properties de mapeamento entre operadores e tags
	 */
	private static ResourceBundle elOpMap =
		ResourceBundle.getBundle(ELEMENT_OPERATOR_MAP);

	/**
	 * Construtor que recebe um elemento raiz, abaixo do qual o filtro ser lido
	 * @param element pai de todos os ns do filtro
	 */
	public DOMFilterBuilder(Element element) {
		this.root = element;
	}

	/**
	 * Cria o objeto filtro representativo das informaes contidas no XML
	 * @return filtro criado
	 */
	public Filter createFilter() {
		if (!this.root.getNodeName().equals("Filter")) {
			LOGGER.warn("Tag Filter esperada. Leu " + this.root.getNodeName());
			return null;
		}
		//itera nos filhos do n raiz, procurando por n que seja do tipo Element
		NodeList nodeList = this.root.getChildNodes();
		Node node = null;
		for (int i = 0; i < nodeList.getLength(); i++) {
			node = nodeList.item(i);
			if (node instanceof Element) {
				break;
			}
		}
		//leia o filtro e retorne
		BooleanValue bv;
		try {
			bv = (BooleanValue) readOperation((Element) node);
		} catch (Exception e) {
			LOGGER.error("Erro na criacao do filtro!", e);
			return null;
		}
		Filter f = new Filter(bv);
		LOGGER.info("Leu filtro: " + f);
		return f;
	}
	/**
	 * dado um nome de tag (Element), instancia a classe correspondente ao
	 * operador
	 * @param elementName nome da tag
	 * @return classe correspondente ao operador
	 */
	private static Class getClassByElementName(String elementName) {
		try {
			String className = elOpMap.getString(elementName);
			return Class.forName(className);
		} catch (ClassNotFoundException e) {
			return null;
		} catch (MissingResourceException e) {
			return null;
		}
	}
	/**
	* Retorna o valor de um element
	* @param el element
	* @return valor do element
	*/
	private static String getElementValue(Element el) {
		NodeList list = el.getChildNodes();
		StringBuffer buf = new StringBuffer();
		//	itere nos filhos do elemento at encontrar ns que implementem Text
		for (int i = 0; i < list.getLength(); i++) {
			Node node = list.item(i) ;
			if (node instanceof Text) {
				//	Encontrado o n de texto, recupere o seu valor
				Text txtNode = (Text) node;
				buf.append(txtNode.getData());
			} 
		}
		
		if (buf.length() == 0) {
			LOGGER.warn("Nenhum valor encontrado para " + el.getNodeName());
			return null;
		}

		return buf.toString();
	}
	
	
	
	/**
	 * Dado um array de objetos e uma classe, encontra o construtor mais
	 * apropriado, de acordo com o tipo dos parametros
	 * @param cls classe (Operator)
	 * @param params tipos dos parametros (Values)
	 * @return objeto de reflexao que representa o construtor encontrado
	 */
	private static Constructor getApropriatedConstructor(
		Class cls,
		Object[] params) {
		//recupere os construtores publicos da classe
		Constructor[] cons = cls.getConstructors();
		for (int i = 0; i < cons.length; i++) {
			//recupere os tipos dos parametros esperados pelo construtor
			Class[] paramTypes = cons[i].getParameterTypes();
			//verifica se o numero de parametros casa com o construtor
			if (paramTypes.length == params.length) {
				boolean accepts = true;
				//verifique se o tipo dos parametros do construtor casam
				//com os tipos fornecidos
				for (int j = 0; j < paramTypes.length; j++) {
					if (!paramTypes[j].isInstance(params[j])) {
						//caso nao case
						accepts = false;
						break;
					}
				}
				if (accepts) {
					//se accpets == true, o construtor foi encontrado
					return cons[i];
				}
			}
		}

		LOGGER.warn(
			"Nenhum construtor adequado localizado para "
				+ cls
				+ arrayToString(params));
		return null;

	}
	private static String arrayToString(Object[] obj) {
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < obj.length; i++) {
			if (i != 0) {
				buf.append(", ");
			}
			buf.append(obj[i]);
		}
		return buf.toString();
	}

	private static Operator readOperation(Element element) throws Exception {
		String nodeName = element.getNodeName();
		Class cls = getClassByElementName(nodeName);
		if (cls == null) {
			LOGGER.error("Operador " + nodeName + "  desconhecido");
			throw new Exception("Operador " + nodeName + "  desconhecido.");
		}
		Object[] param;
		if (cls.isAssignableFrom(LiteralOperator.class)) {
			String value = getElementValue(element);
			param = new Object[] { value };
		} else {
			NodeList list = element.getChildNodes();
			List paramList = new ArrayList();
			for (int i = 0; i < list.getLength(); i++) {
				Node node = list.item(i);
				if (node instanceof Element) {
					Operator aop = readOperation((Element) node);
					if (paramList.size() == 2) {
						String msg = "Operador " + nodeName + " nao suporta mais do que 2 parametros!";
						LOGGER.error(msg);
						throw new Exception(msg);
					}
					paramList.add(aop);
				}
			}
			param = paramList.toArray();
		}
		Constructor cons = getApropriatedConstructor(cls, param);
		if (cons != null) {
			return (Operator) cons.newInstance(param);
		}
		return null;
	}
	
	public static void main(String[] args) throws Exception {
		Filter f = null;
		DocumentBuilderFactory fac = DocumentBuilderFactory.newInstance();
		DocumentBuilder b = fac.newDocumentBuilder();
		InputStream xmlStream =
			DOMFilterBuilder.class.getResourceAsStream("/TesteFilter.xml");
		Document doc = b.parse(xmlStream);
		NodeList list = doc.getElementsByTagName("Filter");
		for (int i = 0; i < list.getLength(); i++) {
			FilterBuilder fb = new DOMFilterBuilder((Element) list.item(0));
			f = fb.createFilter();
		}
		System.out.println(f.toString());

	}

}
